﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.BioBoost.Data {
  [StorableClass]
  [Item("Product", "Summarizes product information.")]
  public class Product : DataItem {

    #region Parameters
    public IValueParameter<StringValue> LabelParameter { get { return (IValueParameter<StringValue>)Parameters["Label"]; } }
    public IValueParameter<DoubleValue> PriceParameter { get { return (IValueParameter<DoubleValue>)Parameters["Price"]; } }
    public IValueParameter<DoubleValue> PriceAtSourceParameter { get { return (IValueParameter<DoubleValue>)Parameters["PriceAtSource"]; } }
    public IValueParameter<DoubleValue> PriceAtTargetParameter { get { return (IValueParameter<DoubleValue>)Parameters["PriceAtTarget"]; } }
    public IValueParameter<DoubleValue> MaxPenaltyParameter { get { return (IValueParameter<DoubleValue>)Parameters["MaxPenalty"]; } }
    public IValueParameter<DoubleValue> PercentageParameter { get { return (IValueParameter<DoubleValue>)Parameters["Percentage"]; } }
    public IValueParameter<DoubleValue> MaxPriceParameter { get { return (IValueParameter<DoubleValue>)Parameters["MaxPrice"]; } }
    public IValueParameter<ValueParameterCollection> PriceScalingParameter { get { return (IValueParameter<ValueParameterCollection>)Parameters["PriceScaling"]; } }
    public IValueParameter<ValueParameterCollection> MaxPriceScalingParameter { get { return (IValueParameter<ValueParameterCollection>)Parameters["MaxPriceScaling"]; } }
    #endregion

    #region Parameter Values
    public string Label { get { return LabelParameter.Value.Value; } }
    public double Price { get { return PriceParameter.Value.Value; } }
    public double PriceAtSource { get { return (PriceAtSourceParameter.Value ?? PriceParameter.Value).Value; } }
    public double PriceAtTarget { get { return (PriceAtTargetParameter.Value ?? PriceParameter.Value).Value; } }
    public double MaxPenalty { get { return MaxPenaltyParameter.Value.Value; } }
    public double Percentage { get { return PercentageParameter.Value.Value; } }
    public double MaxPrice { get { return MaxPriceParameter.Value.Value; } }
    public ValueParameterCollection PriceScaling { get { return PriceScalingParameter.Value; } }
    public ValueParameterCollection MaxPriceScaling { get { return MaxPriceScalingParameter.Value; } }
    #endregion

    #region Construction & Cloning
    public Product() {
      Parameters.Add(new ValueParameter<StringValue>("Label", "The label/name of this product."));
      Parameters.Add(new ValueParameter<DoubleValue>("Price", "The product's price; can be negative to indicate disposal cost.", new DoubleValue(0)));
      Parameters.Add(new OptionalValueParameter<DoubleValue>("PriceAtSource", "The product's price at the source/production location."));
      Parameters.Add(new OptionalValueParameter<DoubleValue>("PriceAtTarget", "The product's price when delivered to the customer."));
      Parameters.Add(new ValueParameter<DoubleValue>("MaxPenalty", "The maximum penalty factor that is then multiplied with the base price at maximum market saturation.", new DoubleValue(1)));
      Parameters.Add(new ValueParameter<DoubleValue>("Percentage", "The percentage of the feedstock utilization where the price starts to increase.", new DoubleValue(0.5)));
      Parameters.Add(new ValueParameter<DoubleValue>("MaxPrice", "The maximum product's price at 100 percent utilization.", new DoubleValue(0)));
      Parameters.Add(new ValueParameter<ValueParameterCollection>("PriceScaling", "Regional scaling of base price", new ValueParameterCollection()));
      Parameters.Add(new ValueParameter<ValueParameterCollection>("MaxPriceScaling", "Regional scaling of max price", new ValueParameterCollection()));
      RegisterEventHandlers();
    }

    [StorableConstructor]
    protected Product(bool isDeserializing) : base(isDeserializing) { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
      if (!Parameters.ContainsKey("PriceScaling"))
        Parameters.Add(new ValueParameter<ValueParameterCollection>("PriceScaling", "Regional scaling of base price", new ValueParameterCollection()));
      if (!Parameters.ContainsKey("MaxPriceScaling"))
        Parameters.Add(new ValueParameter<ValueParameterCollection>("MaxPriceScaling", "Regional scaling of max price", new ValueParameterCollection()));
    }

    protected Product(Product orig, Cloner cloner)
      : base(orig, cloner) {
      RegisterEventHandlers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Product(this, cloner);
    }
    #endregion

    public override void CollectParameterValues(System.Collections.Generic.IDictionary<string, IItem> values) {
      var tmpDict = new Dictionary<string, IItem>();
      base.CollectParameterValues(tmpDict);
      foreach (var entry in tmpDict) {
        if (entry.Key == "Label") continue;
        values.Add("Product " + Label + "." + entry.Key, entry.Value);
      }
    }
    private void RegisterEventHandlers() {
      // when the HL value is changed
      LabelParameter.ValueChanged += (sender, args) => {
        UpdateName();
        LabelParameter.Value.ValueChanged += (s, a) => { UpdateName(); };
      };
      // when the HL value's value is changed
      LabelParameter.Value.ValueChanged += (sender, args) => { UpdateName(); };
    }

    private void UpdateName() {
      this.Name = ItemName + ": " + Label;
    }

    public override bool IsEquivalentTo(DataItem other) {
      var p = other as Product;
      if (p == null) return false;
      return p.Label == Label;
    }

    public double GetRegionalBasePrice(string regionName) {
      IValueParameter priceParam = null;
      IValueParameter<DoubleValue> price;
      while (regionName.Length >= 2) {
        if (PriceScaling.TryGetValue(regionName, out priceParam)) break;
        regionName = regionName.Remove(regionName.Length - 1);
      }
      price = priceParam as IValueParameter<DoubleValue>;
      if (price != null) return Price * price.Value.Value;
      return Price;
    }

    public double GetRegionalMaxPrice(string regionName) {
      IValueParameter maxPriceParam = null;
      IValueParameter<DoubleValue> maxPrice;
      while (regionName.Length >= 2) {
        if (MaxPriceScaling.TryGetValue(regionName, out maxPriceParam)) break;
        regionName = regionName.Remove(regionName.Length - 1);
      }
      maxPrice = maxPriceParam as IValueParameter<DoubleValue>;
      if (maxPrice != null) return MaxPrice * maxPrice.Value.Value;
      return MaxPrice;
    }
  }
}
